// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "landscape.h"
#include "jpeg_loader.h"
#include "quadtree_preprocess.h"
#include "frustum.h"
#include "get_distance.h"
#include <string>
#include <sstream>
#include <iostream>
#include <algorithm>

extern int g_landscape_detail;

//! Constructor. Loads the landscape from files.
Landscape::Landscape(std::string heightfield_filename, std::string texture_filename,
    std::string detail_texture_filename, std::string outer_texture_filename, IGraphicsSystem *igs)
{
    // Store the renderer interface pointer.
    this->IGS = igs;

    // ----------------------------------------------------------------------------------
    // Load terrain into an array of floats.
    // ----------------------------------------------------------------------------------

    heightfield = new float[513 * 513];
    
    FILE *terrain_file;
    terrain_file = fopen(heightfield_filename.c_str(), "rb");
    if (terrain_file == NULL)
    {
        std::stringstream error;
        error << "Failed to load " << heightfield_filename;
        throw error.str();
    }

    fread(heightfield, sizeof(float), 513 * 513, terrain_file);
    fclose(terrain_file);

    // ------------------------------------------------------------------------------------------------
    // Calculate the table of precalculated quadtree vertex errors
    // ------------------------------------------------------------------------------------------------

    precalcerror = quadtree_preprocess(heightfield);

    // ----------------------------------------------------------------------------------
    // Alocate memory for the 2048 x 2048 jpeg
    // ----------------------------------------------------------------------------------

    image = new unsigned char[2048 * 2048 * 3];

    load_jpeg(texture_filename, image);

    // Create a new temporary 256x256 texture
    unsigned char *texture = new unsigned char[256 * 256 * 3];

    for (int t = 0; t < 64; t++)
    {
        // Copy a portion of the larger texture to the new texture
        int left    = (t % 8) * 256;
        int top     = (t / 8) * 256;

        for (int y = 0; y < 256; y++)
        {
/*
            for (int x = 0; x < 256; x++)
            {
                texture[(x + y * 256) * 3 + 0] = image[(x + left + (y + top) * 2048) * 3 + 0];
                texture[(x + y * 256) * 3 + 1] = image[(x + left + (y + top) * 2048) * 3 + 1];
                texture[(x + y * 256) * 3 + 2] = image[(x + left + (y + top) * 2048) * 3 + 2];
            }
*/
            memcpy(texture + (y * 256) * 3, image + (left + (y + top) * 2048) * 3, 256 * 3);
        }

        textures[t] = IGS->CreateTexture(texture, 256, 256, true);
    }

    // Delete[] the temporary 256x256 texture
    delete[] texture;

    // load a detail texture. detail texture is used to hide low resolution of 2048x2048 texture
    detail_texture = IGS->CreateTexture(detail_texture_filename, true);
    
    outer_texture = IGS->CreateTexture(outer_texture_filename, true);

    // ----------------------------------------------------------------------------------
    // Alocate memory for the vertex table. This will indicate if
    // vertices are to be enabled or disabled
    // ----------------------------------------------------------------------------------
    vertexfield = new bool[513 * 513];
    vertexfield_levels = new bool[513 * 513 * 4];

    // ----------------------------------------------------------------------------------
    // Calculate the miny and maxy for each of the 256 blocks in the landscape
    // ----------------------------------------------------------------------------------

    int x;
    for (x = 0; x < 16; x++)
    {
        for (int y = 0; y < 16; y++)
        {
            float miny = 256, maxy = 0;

            for (int i = x * 32; i < x * 32 + 33; i++)
            {
                for (int j = y * 32; j < y * 32 + 33; j++)
                {
                    float h = heightfield[(j * 513) + i];

                    if (miny > h) miny = h;
                    if (maxy < h) maxy = h;
                }
            }

            block_miny[(y * 16) + x] = miny;
            block_maxy[(y * 16) + x] = maxy;
        }
    }

    // ----------------------------------------------------------------------------------
    // Landscape Options
    // ----------------------------------------------------------------------------------

    // 1200 to 5000
    //geometric_detail = 2400;
    geometric_detail = g_landscape_detail;

    // ----------------------------------------------------------------------------------
    // Generate the geomipmaps.
    //
    // Each of the 256 blocks in the landscape has 4 geomipmaps associated with it,
    // representing different detail levels. Here we compute all the geomipmaps for each
    // block. Later when the terrain is modified/deformed the geomipmaps for those blocks
    // must be recalculated.
    // ----------------------------------------------------------------------------------

    for (int level = 0; level < 4; level++)
    {
        // local_detail = 32, 64, 128, 256.
        local_detail = int(pow(2, level + 5));

        // Clear the vertexfield
        memset(vertexfield, 0, sizeof(bool) * 513 * 513);

        for (x=0;x<16;x++)
        {
            for (int y=0;y<16;y++)
            {
                update_block(x,y);
            }
        }

        for (x=0;x<16;x++)
        {
            for (int y=0;y<16;y++)
            {
                dependency_update(x*32,y*32,0,0,32,32);
            }
        }

        for (x=0;x<16;x++)
        {
            for (int y=0;y<16;y++)
            {
                dependency_update(x*32,y*32,0,0,32,32);
            }
        }

        for (x = 0; x < 513; x++)
        {
            for (int y = 0; y < 513; y++)
            {
                vertexfield_levels[x + y * 513 + (level * 513 * 513)] = *(vertexfield + x + y * 513);
            }
        }

        for (x=0;x<16;x++)
        {
            for (int y=0;y<16;y++)
            {
                // --------------------------------------------------------------------
                // The name draw_block is a bit of a misnomer as it doesn't draw
                // anything.
                // It dates from a time when the engine did a lot of CPU intensive LOD
                // per frame. Now I use geomipmapping & static vertex buffers. It's
                // still used when generating the vertex buffers because it does a great
                // job of reducing flat areas of land to few triangles.
                // --------------------------------------------------------------------
                draw_block(x,y);

                // Create a new indexbuffer and vertexbuffer
                //Indices, indexcount, Vertices, vertexcount

                std::vector<unsigned short> ibuffer;
                ibuffer.resize(indexcount);

                int n;
                for (n = 0; n < indexcount; n++)
                {
                    ibuffer[n] = Indices[n];
                }

                segments_indices[x + y * 16 + level * 256] = IGS->CreateIndexBuffer(ibuffer);
                segments_indices_count[x + y * 16 + level * 256] = indexcount;

                VERTEX_XYZ_UV1_UV2* vbuffer = new VERTEX_XYZ_UV1_UV2[vertexcount];

                for (n = 0; n < vertexcount; n++)
                {
                    vbuffer[n].u1 = Vertices[n].u1;
                    vbuffer[n].v1 = Vertices[n].v1;
                    vbuffer[n].u2 = Vertices[n].u2;
                    vbuffer[n].v2 = Vertices[n].v2;
                    vbuffer[n].point.x = Vertices[n].point.x;
                    vbuffer[n].point.y = Vertices[n].point.y;
                    vbuffer[n].point.z = Vertices[n].point.z;
                }

                segments_vertices[x + y * 16 + level * 256]
                    = IGS->CreateVertexList(VTX_XYZ | VTX_UV1 | VTX_UV2, (float*)vbuffer, vertexcount);
                segments_vertices_count[x + y * 16 + level * 256] = vertexcount;
                delete[] vbuffer;
            }
        }
    }
}

//! Destructor
Landscape::~Landscape()
{
    // Delete all the textures
    for (int t = 0; t < 64; t++)
    {
        IGS->DeleteTexture(textures[t]);
    }
    IGS->DeleteTexture(detail_texture);

    for (int v = 0; v < 256; v++)
    {
        for (int level = 0; level < 4; level++)
        {
            IGS->DeleteVertexList(segments_vertices[v + level * 256]);
            IGS->DeleteIndexBuffer(segments_indices[v + level * 256]);
        }
    }

    delete[] vertexfield;
    delete[] vertexfield_levels;
    delete[] precalcerror;
    delete[] heightfield;   // Delete the float 513x513 heightfield.
    delete[] image;
}

//! Update function. Call this no more than once per frame.
void Landscape::Update(vector3 camera)
{
    for (int x=0;x<16;x++)
    {
        for (int y=0;y<16;y++)
        {
            float aabb[6];

            aabb[0] = (float)x * 32.0f;
            aabb[1] = (float)x * 32.0f + 32.0f;
            aabb[2] = block_miny[(y * 16) + x];
            aabb[3] = block_maxy[(y * 16) + x];
            aabb[4] = (float)y * 32.0f;
            aabb[5] = (float)y * 32.0f + 32.0f;

            if (AABBInFrustum(aabb))
            {
                block_visible[x + y * 16] = true;

                float pos_x = (float)x * 32.0f + 16.0f;
                float pos_z = (float)y * 32.0f + 16.0f;
                float pos_y = *(heightfield + ((x * 32) + 16) + ((y * 32) + 16) * 513);

                float length = get_distance(camera, pos_x, pos_y, pos_z);

                block_detail[x + y * 16] = 0;
                if (length > 64)
                    block_detail[x + y * 16] = 1;
                if (length > 128)
                    block_detail[x + y * 16] = 2;
                if (length > 256)
                    block_detail[x + y * 16] = 3;
            }
            else block_visible[x + y * 16] = false;
        }
    }

}

class BlockDepthInfo
{
public:
    int x, y;
    float distance_from_camera;

    // comparison operator. this will be called by std::sort and will sort them by distance
    bool operator<(const BlockDepthInfo& other) const
    {
        return (this->distance_from_camera < other.distance_from_camera);
    }
};

//! Render function. Call this once per frame.
void Landscape::Render(vector3 camera)
{
    IGS->EnableLighting(false);
    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, IdentityMatrix44());

    std::vector<BlockDepthInfo> depths;
    depths.resize(256);

    for (int x=0;x<16; x++)
    {
        for (int y=0;y<16;y++)
        {
            float pos_x = (float)x * 32.0f + 16.0f;
            float pos_z = (float)y * 32.0f + 16.0f;
            float pos_y = *(heightfield + ((x * 32) + 16) + ((y * 32) + 16) * 513);

            depths[x + y * 16].distance_from_camera = get_distance(camera, pos_x, pos_y, pos_z);
            depths[x + y * 16].x = x;
            depths[x + y * 16].y = y;
        }
    }

    // Sort the landscape blocks by distance
    std::sort(depths.begin(), depths.end());

    // Loop through all the landscape blocks and render them
    std::vector<BlockDepthInfo>::iterator itor;
    for (itor = depths.begin(); itor != depths.end(); itor++)
    {
        int x = itor->x;
        int y = itor->y;

        if (block_visible[x + y * 16])
        {
            IGS->SetMultiTextureMode(DETAIL_TEXTURE);
            IGS->SetTexture(1, textures[(x / 2) + (y / 2) * 8]);
            IGS->SetTextureWrappingMode(1, CLAMP_TEXTURE);

            IGS->SetTexture(0, detail_texture);
            IGS->SetTextureWrappingMode(0, WRAP_TEXTURE);

            IGS->SetCullMode(CULL_NONE);

            int level = block_detail[x + y * 16];

            if (y < 15 && block_visible[x + (y + 1) * 16])
            {
                if (level != block_detail[x + (y + 1) * 16])
                {
                    draw_edge_strip(x, y, x, y + 1);
                }
            }
            if (x < 15 && block_visible[(x + 1) + y * 16])
            {
                if (level != block_detail[(x + 1) + y * 16])
                {
                    draw_edge_strip(x, y, x + 1, y);
                }
            }

            IGS->SetCullMode(CULL_CCW);

            IGS->DrawIndexedTriangles(TRIANGLE_LIST,
                segments_vertices[x + y * 16 + level * 256],
                segments_vertices_count[x + y * 16 + level * 256],
                segments_indices[x + y * 16 + level * 256],
                segments_indices_count[x + y * 16 + level * 256]);
        }

    }
/*
    for (int x=0;x<16;x++)
    {
        for (int y=0;y<16;y++)
        {
            if (block_visible[x + y * 16])
            {
                IGS->SetMultiTextureMode(DETAIL_TEXTURE);
                IGS->SetTexture(1, textures[(x / 2) + (y / 2) * 8]);
                IGS->SetTextureWrappingMode(1, CLAMP_TEXTURE);

                IGS->SetTexture(0, detail_texture);
                IGS->SetTextureWrappingMode(0, WRAP_TEXTURE);

                IGS->SetCullMode(CULL_NONE);

                int level = block_detail[x + y * 16];

                if (y < 15 && block_visible[x + (y + 1) * 16])
                {
                    if (level != block_detail[x + (y + 1) * 16])
                    {
                        draw_edge_strip(x, y, x, y + 1);
                    }
                }
                if (x < 15 && block_visible[(x + 1) + y * 16])
                {
                    if (level != block_detail[(x + 1) + y * 16])
                    {
                        draw_edge_strip(x, y, x + 1, y);
                    }
                }

                IGS->SetCullMode(CULL_CCW);

                IGS->DrawIndexedTriangles(TRIANGLE_LIST,
                    segments_vertices[x + y * 16 + level * 256],
                    segments_vertices_count[x + y * 16 + level * 256],
                    segments_indices[x + y * 16 + level * 256],
                    segments_indices_count[x + y * 16 + level * 256]);
            }
        }
    }
    */

    IGS->SetTexture(1, 0);
    IGS->SetTexture(0, 0);

    IGS->SetMultiTextureMode(NORMAL);
    IGS->SetTexture(0, outer_texture);
    IGS->SetTextureWrappingMode(0, WRAP_TEXTURE);

    vector3 land_surface[4];

    land_surface[0] = vector3(-1.0f,  0.0f,  1.0f);
    land_surface[1] = vector3(-1.0f,  0.0f, -1.0f);
    land_surface[2] = vector3( 1.0f,  0.0f, -1.0f);
    land_surface[3] = vector3( 1.0f,  0.0f,  1.0f);

    float scale = 256.0f;

    for (int n = 0; n < 4; n++)
    {
        land_surface[n].x *= scale;
        land_surface[n].y *= scale;
        land_surface[n].z *= scale;

        land_surface[n].x += 256.0f;
        land_surface[n].z += 256.0f;

        land_surface[n] -= vector3(2048.0f, 0.05f, 2048.0f);
    }

    VERTEX_XYZ_UV1 points[4];

    points[0].u = 0.0f;
    points[0].v = 0.0f;

    points[1].u = 0.0f;
    points[1].v = 16.0f;

    points[2].u = 16.0f;
    points[2].v = 16.0f;

    points[3].u = 16.0f;
    points[3].v = 0.0f;

    unsigned short indices[] = { 0, 1, 2, 2, 3, 0};

    for (int i = 0; i < 10; i++)
    {
        for (int j = 0; j < 10; j++)
        {
            if (i != 4 || j != 4)
            {
                points[0].point = land_surface[0] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
                points[1].point = land_surface[1] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
                points[2].point = land_surface[2] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
                points[3].point = land_surface[3] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);

                IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_UV1, (float*)points, 4, indices, 6);
            }
        }
    }
}

//! Terrain Update.
void Landscape::UpdateTerrainRectangle(update_rect& rect)
{
    int left    = rect.left;
    int right   = rect.right;
    int top     = rect.top;
    int bottom  = rect.bottom;

    if (left > right || top > bottom)   return;
    if (left == right || top == bottom) return;

    int xblock, yblock;

    // -------------------------------------------------------------------------
    // Recalculate the quadtree lookup table for the blocks this rectangle
    // overlaps with.
    // -------------------------------------------------------------------------
    for (xblock = left/32; xblock <= right/32; xblock++)
    {
        for (yblock = top/32; yblock <= bottom/32; yblock++)
        {
            if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
            {
                quadtree_reprocess(heightfield, precalcerror, xblock, yblock);

                // ----------------------------------------------------------------------------------
                // Calculate the miny and maxy for the AABBs of each of the 256 blocks in the landscape
                // These are used when frustum culling.
                // ----------------------------------------------------------------------------------
                float miny = 256, maxy = 0;

                for (int i = xblock * 32; i < xblock * 32 + 33; i++)
                {
                    for (int j = yblock * 32; j < yblock * 32 + 33; j++)
                    {
                        float h = heightfield[(j * 513) + i];

                        if (miny > h) miny = h;
                        if (maxy < h) maxy = h;
                    }
                }

                block_miny[(yblock * 16) + xblock] = miny;
                block_maxy[(yblock * 16) + xblock] = maxy;
            }
        }
    }

    // -------------------------------------------------------------------------
    // Regenerate the vertex buffers for all of the geomipmap levels of blocks
    // directly affected by the deformation and also adjacent blocks.
    // -------------------------------------------------------------------------

    for (int level = 0; level < 4; level++)
    {
        // local_detail = 32, 64, 128, 256.
        local_detail = int(pow(2, level + 5));

        // Clear the vertexfield
        memset(vertexfield, 0, sizeof(bool) * 513 * 513);

        for (xblock = (left - 48)/32; xblock <= (right+48)/32; xblock++)
        {
            for (yblock = (top - 48)/32; yblock <= (bottom+48)/32; yblock++)
            {
                if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
                {
                    update_block(xblock, yblock);
                }
            }
        }

        for (xblock = (left - 48)/32; xblock <= (right+48)/32; xblock++)
        {
            for (yblock = (top - 48)/32; yblock <= (bottom+48)/32; yblock++)
            {
                if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
                {
                    dependency_update(xblock*32, yblock*32, 0, 0, 32, 32);
                }
            }
        }

        for (xblock = (left - 48)/32; xblock <= (right+48)/32; xblock++)
        {
            for (yblock = (top - 48)/32; yblock <= (bottom+48)/32; yblock++)
            {
                if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
                {
                    dependency_update(xblock*32,yblock*32,0,0,32,32);
                }
            }
        }

        for (xblock = (left - 16)/32; xblock <= (right+16)/32; xblock++)
        {
            for (yblock = (top - 16)/32; yblock <= (bottom+16)/32; yblock++)
            {
                if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
                {
                    for (int x = 0; x < 33; x ++)
                    {
                        for (int y = 0; y < 33; y++)
                        {
                            int x_local = xblock * 32 + x;
                            int y_local = yblock * 32 + y;
                            vertexfield_levels[x_local + y_local * 513 + (level * 513 * 513)] =
                                *(vertexfield + x_local + y_local * 513);
                        }
                    }
                }
            }
        }

        for (xblock = (left - 16)/32; xblock <= (right+16)/32; xblock++)
        {
            for (yblock = (top - 16)/32; yblock <= (bottom+16)/32; yblock++)
            {
                if (xblock >= 0 && yblock >= 0 && xblock < 16 && yblock < 16)
                {
                    // --------------------------------------------------------------------
                    // Delete the old vertex list and index list
                    // --------------------------------------------------------------------
                    IGS->DeleteVertexList(segments_vertices[xblock + yblock * 16 + level * 256]);
                    IGS->DeleteIndexBuffer(segments_indices[xblock + yblock * 16 + level * 256]);

                    // --------------------------------------------------------------------
                    // The name draw_block is a bit of a misnomer as it doesn't draw
                    // anything.
                    // It dates from a time when the engine did a lot of CPU intensive LOD
                    // per frame. Now I use geomipmapping & static vertex buffers. It's
                    // still used when generating the vertex buffers because it does a great
                    // job of reducing flat areas of land to few triangles.
                    // --------------------------------------------------------------------
                    draw_block(xblock,yblock);

                    // Create a new indexbuffer and vertexbuffer
                    //Indices, indexcount, Vertices, vertexcount

                    std::vector<unsigned short> ibuffer;
                    ibuffer.resize(indexcount);

                    int n;
                    for (n = 0; n < indexcount; n++)
                    {
                        ibuffer[n] = Indices[n];
                    }

                    segments_indices[xblock + yblock * 16 + level * 256] = IGS->CreateIndexBuffer(ibuffer);
                    segments_indices_count[xblock + yblock * 16 + level * 256] = indexcount;

                    VERTEX_XYZ_UV1_UV2* vbuffer = new VERTEX_XYZ_UV1_UV2[vertexcount];

                    for (n = 0; n < vertexcount; n++)
                    {
                        vbuffer[n].u1 = Vertices[n].u1;
                        vbuffer[n].v1 = Vertices[n].v1;
                        vbuffer[n].u2 = Vertices[n].u2;
                        vbuffer[n].v2 = Vertices[n].v2;
                        vbuffer[n].point.x = Vertices[n].point.x;
                        vbuffer[n].point.y = Vertices[n].point.y;
                        vbuffer[n].point.z = Vertices[n].point.z;
                    }

                    segments_vertices[xblock + yblock * 16 + level * 256]
                        = IGS->CreateVertexList(VTX_XYZ | VTX_UV1 | VTX_UV2, (float*)vbuffer, vertexcount);
                    segments_vertices_count[xblock + yblock * 16 + level * 256] = vertexcount;
                    delete[] vbuffer;
                }
            }
        }
    }
}

//! Terrain texture update
void Landscape::UpdateTerrainTexture(int width, int height, int x, int y, unsigned char *tex_buffer)
{
    for (int j = y; j < y + height; j++)
    {
        for (int i = x; i < x + width; i++)
        {
            if (i >= 0 && j >= 0 && x < 2048 && y < 2048)
            {
                // pixel in 2048x2048 texture = image[(i + j * 2048) * 3]
                // pixel in buffer = tex_buffer[((i - x) + (j - y) * width) * 4]

                unsigned char r1 = image[((i + j * 2048) * 3) + 0];
                unsigned char g1 = image[((i + j * 2048) * 3) + 1];
                unsigned char b1 = image[((i + j * 2048) * 3) + 2];

                unsigned char a2 = tex_buffer[(((i - x) + (j - y) * width) * 4) + 0];
                unsigned char r2 = tex_buffer[(((i - x) + (j - y) * width) * 4) + 1];
                unsigned char g2 = tex_buffer[(((i - x) + (j - y) * width) * 4) + 2];
                unsigned char b2 = tex_buffer[(((i - x) + (j - y) * width) * 4) + 3];

                unsigned char r3 = (r1 * (255 - a2) + r2 * a2) / 255;
                unsigned char g3 = (g1 * (255 - a2) + g2 * a2) / 255;
                unsigned char b3 = (b1 * (255 - a2) + b2 * a2) / 255;

                image[((i + j * 2048) * 3) + 0] = r3;
                image[((i + j * 2048) * 3) + 1] = g3;
                image[((i + j * 2048) * 3) + 2] = b3;
            }
        }
    }


    // Check which of the 64 textures need to be updated
    int left    =       x       / 256;
    int right   = (x + width)   / 256;
    int top     =       y       / 256;
    int bottom  = (y + height)  / 256;

    // Create a new temporary 256x256 texture
    unsigned char *texture = new unsigned char[256 * 256 * 3];

    for (int i2 = left; i2 <= right; i2++)
    {
        for (int j2 = top; j2 <= bottom; j2++)
        {
            if (i2 >= 0 && j2 >= 0 && i2 < 8 && j2 < 8)
            {
                int t = i2 + j2 * 8;

                // Copy a portion of the larger texture to the new texture
                int left_offset = i2 * 256;
                int top_offset  = j2 * 256;

                for (int row = 0; row < 256; row++)
                {
                    memcpy(texture + (row * 256) * 3,
                        image + (left_offset + (row + top_offset) * 2048) * 3, 256 * 3);
                }

                // Delete the old texture
                IGS->DeleteTexture(textures[t]);

                // Create a new texture
                textures[t] = IGS->CreateTexture(texture, 256, 256, true);
            }
        }
    }

    // Delete[] the temporary 256x256 texture
    delete[] texture;
}

//! Returns the height of the heightfield at a certain point
float Landscape::GetHeight(float x, float y)
{
    if (x < 0.0f || x > 513.0f || y < 0.0f || y > 513.0f)
    {
        return 0.0f;
    }
    else if (floor(x) == x && floor(y) == y)
    {
        int ix = (int)x;
        int iy = (int)y;

        return *(heightfield + ix + iy * 513);
    }
    else // interpolate to calculate the height
    {
        float x1 = floor(x);
        float y1 = floor(y);
        float x2 = ceil(x);
        float y2 = ceil(y);

        float h1 = GetHeight(x1,y1);
        float h2 = GetHeight(x2,y1);
        float h3 = GetHeight(x1,y2);
        float h4 = GetHeight(x2,y2);

        float h_top = h1 + ((h2 - h1) * (x1 - x));
        float h_bot = h3 + ((h4 - h3) * (x1 - x));

        float h_final = h_top + ((h_bot - h_top) * (y1 - y));

        return h_final;
    }
}

// Given an x,y pair representing 1 of the 256 blocks in the quadtree, and the square within
// that block (left, right, bottom, top) this function will determine from the array of
// precalculated errors, the errors for the 4 subsquares in that square.
void Landscape::geterrors(int x,int y, int left, int top, int right, int bottom,
    int& nw_error, int& ne_error, int& sw_error, int& se_error)
{
    unsigned char* offset = precalcerror + (342 * (x+(y*16)));

    int size = (right-left);

    int i = left / (size/2);
    int j = top  / (size/2);

    if (size==32)
    {
        offset+=2;
        nw_error = *(offset+0);
        ne_error = *(offset+1);
        sw_error = *(offset+2);
        se_error = *(offset+3);
    }
    else if (size==16)
    {
        offset+= (2+4);
        nw_error = *(offset+i  +(j*4 )    );
        ne_error = *(offset+i+1+(j*4 )    );
        sw_error = *(offset+i  +((j+1)*4 ));
        se_error = *(offset+i+1+((j+1)*4 ));
    }
    else if (size==8)
    {
        offset+= (2+4+16);
        nw_error = *(offset+i  +(j*8 )    );
        ne_error = *(offset+i+1+(j*8 )    );
        sw_error = *(offset+i  +((j+1)*8 ));
        se_error = *(offset+i+1+((j+1)*8 ));
    }
    else if (size==4)
    {
        offset+= (2+4+16+64);
        nw_error = *(offset+i  +(j*16)    );
        ne_error = *(offset+i+1+(j*16)    );
        sw_error = *(offset+i  +((j+1)*16));
        se_error = *(offset+i+1+((j+1)*16));
    }
}

// Given an x,y pair representing 1 of the 256 blocks in the landscape, and the square within
// that block (left, right, bottom, top) this function will enable 5 of the vertices within that
// square, namely the centre and the 4 corners.
void Landscape::minimal_update(int x, int y, int left, int top, int right, int bottom)
{
    int centre_x = (left+right)/2;
    int centre_y = (top+bottom)/2;

    enablevertex( x+left   , y+top    );
    enablevertex( x+right  , y+top    );
    enablevertex( x+left   , y+bottom );
    enablevertex( x+right  , y+bottom );

    enablevertex( x+centre_x, y+centre_y );
}

// Given an x,y pair representing 1 of the 256 blocks in the landscape, and the square within
// that block (left, right, bottom, top) this function will recursively enable some of the
// vertices of its subsquares in an attempt to get rid of some T-junctions which cause gaps
// between polygons.
void Landscape::dependency_update(int x, int y, int left, int top, int right, int bottom)
{
    int centre_x = (left+right)/2;
    int centre_y = (top+bottom)/2;

    if (isvertexenabled(x+centre_x,y+top)   || isvertexenabled(x+left,y+centre_y) ||
        isvertexenabled(x+right,y+centre_y) || isvertexenabled(x+centre_x,y+bottom))
    {
        minimal_update(x, y, left, top, right, bottom);
    }

    if ( (right-left)>4)
    {
        dependency_update(x,y,left,top,centre_x,centre_y);
        dependency_update(x,y,centre_x,top,right,centre_y);
        dependency_update(x,y,left,centre_y,centre_x,bottom);
        dependency_update(x,y,centre_x,centre_y,right,bottom);
    }
}

// This function calculates the precalculated error of each of its subsquares and divides that
// by the distance from the camera and selectively updates those subsquares.
void Landscape::recursive_update(int x, int y, int left, int top, int right, int bottom)
{
    int centre_x = (left+right)/2;
    int centre_y = (top+bottom)/2;

    int length;

/*
    float xloc = x + centre_x;
    float zloc = y + centre_y;
    float yloc = *(heightfield + x+ centre_x + (y + centre_y) * 513);
    //length = (int)getdistance(cam_x, cam_y, cam_z, xloc, yloc, zloc);
*/
    length = local_detail;

    int nw_error, ne_error, sw_error, se_error;

    geterrors(x/32, y/32, left, top, right, bottom, nw_error, ne_error, sw_error, se_error);

    int size=right-left;

    nw_error *= geometric_detail;
    ne_error *= geometric_detail;
    sw_error *= geometric_detail;
    se_error *= geometric_detail;

    if (length == 0) length = 1;    // Avoid division by zero
    nw_error /= length;
    ne_error /= length;
    sw_error /= length;
    se_error /= length;

    enablevertex( x+left   , y+top    );
    enablevertex( x+right  , y+top    );
    enablevertex( x+left   , y+bottom );
    enablevertex( x+right  , y+bottom );

    enablevertex( x+centre_x, y+centre_y );

    if ( size > 2 )
    {
        if (nw_error>100)
        {
            recursive_update(x,y, left,top,centre_x,centre_y);

            if ( (x+left-size)>0)
            {
                enablevertex( x+left-size   , y+top    );
                enablevertex( x+right-size  , y+top    );
                enablevertex( x+left -size  , y+bottom );
                enablevertex( x+right -size , y+bottom );

                enablevertex( x+centre_x-size, y+centre_y );
            }
            if ( (y+top -size)>0)
            {

                enablevertex( x+left   , y+top   -size );
                enablevertex( x+right  , y+top   -size );
                enablevertex( x+left  , y+bottom -size);
                enablevertex( x+right  , y+bottom -size);

                enablevertex( x+centre_x, y+centre_y -size);
            }
        }
        if (ne_error>100)
        {
            recursive_update (x,y, centre_x,top,right,centre_y);

            if ( (x+right +size)<513 )
            {
                enablevertex( x+left+size   , y+top    );
                enablevertex( x+right+size  , y+top    );
                enablevertex( x+left +size  , y+bottom );
                enablevertex( x+right +size , y+bottom );

                enablevertex( x+centre_x+size, y+centre_y );
            }
            if ( (y+top   -size)>0)
            {
                enablevertex( x+left   , y+top   -size );
                enablevertex( x+right  , y+top   -size );
                enablevertex( x+left  , y+bottom -size);
                enablevertex( x+right  , y+bottom -size);

                enablevertex( x+centre_x, y+centre_y -size);
            }
        }
        if (sw_error>100)
        {
            recursive_update(x,y, left,centre_y,centre_x,bottom);

            if ( (x+left-size)>0 )
            {
                enablevertex( x+left-size   , y+top    );
                enablevertex( x+right-size  , y+top    );
                enablevertex( x+left -size  , y+bottom );
                enablevertex( x+right -size , y+bottom );

                enablevertex( x+centre_x-size, y+centre_y );
            }
            if ( (y+bottom+size)<513)
            {

                enablevertex( x+left   , y+top +size );
                enablevertex( x+right  , y+top   +size );
                enablevertex( x+left  , y+bottom +size);
                enablevertex( x+right  , y+bottom +size);

                enablevertex( x+centre_x, y+centre_y +size);
            }
        }
        if (se_error>100)
        {
            recursive_update(x,y, centre_x,centre_y,right,bottom);

            if ( (x+right +size)<513 )
            {
                enablevertex( x+left+size   , y+top    );
                enablevertex( x+right+size  , y+top    );
                enablevertex( x+left +size  , y+bottom );
                enablevertex( x+right +size , y+bottom );

                enablevertex( x+centre_x+size, y+centre_y );
            }
            if ( (y+bottom +size)<513)
            {
                enablevertex( x+left   , y+top   +size );
                enablevertex( x+right  , y+top   +size );
                enablevertex( x+left  , y+bottom +size);
                enablevertex( x+right  , y+bottom +size);

                enablevertex( x+centre_x, y+centre_y +size);
            }
        }
    }
    else // (size <= 2)
    {
        enablevertex( x+centre_x   , y+top    );
        enablevertex( x+right  , y+centre_y    );
        enablevertex( x+centre_x   , y+bottom );
        enablevertex( x+left  , y+centre_y );

        if ( (x+right +size)<513 )
        {
            enablevertex( x+left+size   , y+top    );
            enablevertex( x+right+size  , y+top    );
            enablevertex( x+left +size  , y+bottom );
            enablevertex( x+right +size , y+bottom );

            enablevertex( x+centre_x+size, y+centre_y );
        }
        if ( (y+bottom +size)<513)
        {
            enablevertex( x+left   , y+top   +size );
            enablevertex( x+right  , y+top   +size );
            enablevertex( x+left  , y+bottom +size);
            enablevertex( x+right  , y+bottom +size);

            enablevertex( x+centre_x, y+centre_y +size);
        }
        if ( (x+left-size)>0 )
        {
            enablevertex( x+left-size   , y+top    );
            enablevertex( x+right-size  , y+top    );
            enablevertex( x+left -size  , y+bottom );
            enablevertex( x+right -size , y+bottom );

            enablevertex( x+centre_x-size, y+centre_y );
        }
        if ( (y+top -size)>0)
        {
            enablevertex( x+left   , y+top   -size );
            enablevertex( x+right  , y+top   -size );
            enablevertex( x+left  , y+bottom -size);
            enablevertex( x+right  , y+bottom -size);

            enablevertex( x+centre_x, y+centre_y -size);
        }
    }
}

void Landscape::update_block(int x, int y)
{
    recursive_update(x*32, y*32 ,0 ,0 ,32 ,32);
}

void Landscape::recursive_render(int x, int y, int left, int top, int right, int bottom)
{
    if ( (right - left) > 2 )
    {
        // This is not the smallest size

        bool nw_enabled=false;
        bool ne_enabled=false;
        bool sw_enabled=false;
        bool se_enabled=false;

        int centre_x = (left+right)/2;
        int centre_y = (top+bottom)/2;

        int w_centre_x = (left+centre_x)/2;
        int n_centre_y = (top+centre_y)/2;
        int e_centre_x = (right+centre_x)/2;
        int s_centre_y = (bottom+centre_y)/2;

        if (isvertexenabled(x+w_centre_x,y+n_centre_y))
        {
            nw_enabled = true;
            recursive_render(x,y,left,top,centre_x,centre_y);
        }
        if (isvertexenabled(x+e_centre_x,y+n_centre_y))
        {
            ne_enabled = true;
            recursive_render(x,y,centre_x,top,right,centre_y);
        }
        if (isvertexenabled(x+w_centre_x,y+s_centre_y))
        {
            sw_enabled = true;
            recursive_render(x,y,left,centre_y,centre_x,bottom);
        }
        if (isvertexenabled(x+e_centre_x,y+s_centre_y))
        {
            se_enabled = true;
            recursive_render(x,y,centre_x,centre_y,right,bottom);
        }

        bool n_enabled=false;
        bool w_enabled=false;
        bool e_enabled=false;
        bool s_enabled=false;

        if (isvertexenabled(x+centre_x,y+top) )
        {
            n_enabled = true;
        }
        if (isvertexenabled(x+centre_x,y+bottom) )
        {
            s_enabled = true;
        }
        if (isvertexenabled(x+left,y+centre_y) )
        {
            w_enabled = true;
        }
        if (isvertexenabled(x+right,y+centre_y) )
        {
            e_enabled = true;
        }

        // Draw the triangles

        // Do the North bit
        if (ne_enabled && nw_enabled)
        {
            // do nothing
        }
        else if (ne_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
        else if (nw_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
        else if (n_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }

        // Do the South bit
        if (se_enabled && sw_enabled)
        {
            // do nothing
        }
        else if (se_enabled)
        {
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
        }
        else if (sw_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }
        else if (s_enabled)
        {
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }

        // Do the West bit
        if (nw_enabled && sw_enabled)
        {
            // do nothing
        }
        else if (sw_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
        }
        else if (nw_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
        }
        else if (w_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
        }

        // Do the East bit
        if (ne_enabled && se_enabled)
        {
            // do nothing
        }
        else if (se_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
        }
        else if (ne_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }
        else if (e_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
    }
    else // (size <= 2)
    {
        bool n_enabled=false;
        bool w_enabled=false;
        bool e_enabled=false;
        bool s_enabled=false;

        int centre_x = (left+right)/2;
        int centre_y = (top+bottom)/2;

        if (isvertexenabled(x+centre_x,y+top) )
        {
            n_enabled = true;
        }
        if (isvertexenabled(x+centre_x,y+bottom) )
        {
            s_enabled = true;
        }
        if (isvertexenabled(x+left,y+centre_y) )
        {
            w_enabled = true;
        }
        if (isvertexenabled(x+right,y+centre_y) )
        {
            e_enabled = true;
        }

        // Draw the triangles

        // Do the North bit
        if (n_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }

        // Do the South bit
        if (s_enabled)
        {
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
        }

        // Do the West bit
        if (w_enabled)
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[left+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[left+(bottom*33)];
            indexcount++;
        }

        // Do the East bit
        if (e_enabled)
        {
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(centre_y*33)];
            indexcount++;
        }
        else
        {
            Indices[indexcount]=vertextable[right+(top*33)];
            indexcount++;
            Indices[indexcount]=vertextable[right+(bottom*33)];
            indexcount++;
            Indices[indexcount]=vertextable[centre_x+(centre_y*33)];
            indexcount++;
        }
    }
}

void Landscape::draw_block(int i, int j)
{
    indexcount = 0;
    vertexcount= 0;

    bool    *vertexoffset   = vertexfield + i * 32 + (j * 32) * 513;
    float   *heightoffset   = heightfield + i * 32 + (j * 32) * 513;

    float x_float =(float)i*32.0f;
    float z_float =(float)j*32.0f;

    float u1_offset = 0.5f * (i % 2);
    float v1_offset = 0.5f * (j % 2);

    float inv64 = 1.0f / 64.0f;

    for (int y=0;y<33;y++)
    {
        for (int x = 0;x < 33; x++)
        {
            if (*(vertexoffset++)  != false)
            {
                Vertices[vertexcount].point.x = x_float + x;
                Vertices[vertexcount].point.y = *(heightoffset+x);
                Vertices[vertexcount].point.z = z_float + y;

                Vertices[vertexcount].u1 = (u1_offset + x * inv64) * 16.0f;
                Vertices[vertexcount].v1 = (v1_offset + y * inv64) * 16.0f;
                Vertices[vertexcount].u2 = (u1_offset + x * inv64);
                Vertices[vertexcount].v2 = (v1_offset + y * inv64);

                vertextable[x + y * 33] = (unsigned short)vertexcount;
                vertexcount++;
            }
        }

        vertexoffset += (513 - 33);
        heightoffset += 513;
    }

    recursive_render(i*32, j*32, 0, 0, 32, 32);

    //IGS->SetTexture(textures[(i / 2) + (j / 2) * 8]);

    //IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, Vertices, (int)vertexcount, Indices, indexcount);

    //std::cout << "vertexcount = " << vertexcount << " indexcount = " << indexcount << std::endl;
}

void Landscape::draw_edge_strip(int x1, int y1, int x2, int y2)
{
    indexcount = 0;
    vertexcount= 0;

    int x_start, y_start, x_inc, y_inc, vertexoffset_increment;

    float u_inc, v_inc;

    float u_offset, v_offset;

    if (x1 < x2)
    {
        // --- ---
        // |1| |2|
        // --- ---
        x_start = x1 * 32 + 32;
        y_start = y1 * 32;
        x_inc = 0;
        y_inc = 1;

        u_inc = 0.0f;
        v_inc = 1.0f / 64.0f;

        vertexoffset_increment = 513;

        u_offset = 0.5f * (x1 % 2) + 0.5f;
        v_offset = 0.5f * (y1 % 2);
    }
    else // (y1 < y2)
    {
        // ---
        // |1|
        // ---
        // ---
        // |2|
        // ---
        x_start = x1 * 32;
        y_start = y1 * 32 + 32;
        x_inc = 1;
        y_inc = 0;

        u_inc = 1.0f / 64.0f;
        v_inc = 0.0f;

        vertexoffset_increment = 1;

        u_offset = 0.5f * (x1 % 2);
        v_offset = 0.5f * (y1 % 2) + 0.5f;
    }

    // Determine which block is lowest detail
    int level1 = block_detail[x1 + y1 * 16];
    int level2 = block_detail[x2 + y2 * 16];

    bool *vertexoffset_low, *vertexoffset_high;

    if (level1 > level2)
    {
        vertexoffset_low  = (bool *)&vertexfield_levels[x_start + y_start * 513 + (level1 * 513 * 513)];
        vertexoffset_high = (bool *)&vertexfield_levels[x_start + y_start * 513 + (level2 * 513 * 513)];
    }
    else
    {
        vertexoffset_high = (bool *)&vertexfield_levels[x_start + y_start * 513 + (level1 * 513 * 513)];
        vertexoffset_low  = (bool *)&vertexfield_levels[x_start + y_start * 513 + (level2 * 513 * 513)];
    }

    float u2 = 0.0f, v2 = 0.0f;

    VERTEX_XYZ_UV1_UV2 intermediate_vertices[33];

    int intermediate_count = 0;
    for (int n = 0; n < 33; n++)
    {
        float x = (float)x_start;
        float y = *(heightfield + x_start + y_start * 513);
        float z = (float)y_start;

        float u = u_offset + u2;
        float v = v_offset + v2;

        if ((*vertexoffset_low == true) && (*vertexoffset_high == true))
        {
            if (intermediate_count > 1)
            {
                intermediate_vertices[intermediate_count].point.x = x;
                intermediate_vertices[intermediate_count].point.y = y;
                intermediate_vertices[intermediate_count].point.z = z;
                intermediate_vertices[intermediate_count].u1 = u * 16.0f;
                intermediate_vertices[intermediate_count].v1 = v * 16.0f;
                intermediate_vertices[intermediate_count].u2 = u;
                intermediate_vertices[intermediate_count].v2 = v;
                intermediate_count++;

                unsigned short relocate = (unsigned short)vertexcount;

                int v;
                for (v = 0; v < intermediate_count; v++)
                {
                    Vertices[vertexcount].point.x   = intermediate_vertices[v].point.x;
                    Vertices[vertexcount].point.y   = intermediate_vertices[v].point.y;
                    Vertices[vertexcount].point.z   = intermediate_vertices[v].point.z;
                    Vertices[vertexcount].u1        = intermediate_vertices[v].u1;
                    Vertices[vertexcount].v1        = intermediate_vertices[v].v1;
                    Vertices[vertexcount].u2        = intermediate_vertices[v].u2;
                    Vertices[vertexcount].v2        = intermediate_vertices[v].v2;

                    vertexcount++;
                }

                unsigned short final_index  = (unsigned short)intermediate_count - 1;

                for (v = 0; v < intermediate_count - 2; v++)
                {
                    Indices[indexcount + 0] = relocate + v;
                    Indices[indexcount + 1] = relocate + v + 1;
                    Indices[indexcount + 2] = relocate + final_index;
                    indexcount += 3;
                }
            }

            intermediate_count = 0;

            intermediate_vertices[intermediate_count].point.x = x;
            intermediate_vertices[intermediate_count].point.y = y;
            intermediate_vertices[intermediate_count].point.z = z;
            intermediate_vertices[intermediate_count].u1 = u * 16.0f;
            intermediate_vertices[intermediate_count].v1 = v * 16.0f;
            intermediate_vertices[intermediate_count].u2 = u;
            intermediate_vertices[intermediate_count].v2 = v;
            intermediate_count++;
        }
        else if (*vertexoffset_high == true)
        {
            intermediate_vertices[intermediate_count].point.x = x;
            intermediate_vertices[intermediate_count].point.y = y;
            intermediate_vertices[intermediate_count].point.z = z;
            intermediate_vertices[intermediate_count].u1 = u * 16.0f;
            intermediate_vertices[intermediate_count].v1 = v * 16.0f;
            intermediate_vertices[intermediate_count].u2 = u;
            intermediate_vertices[intermediate_count].v2 = v;

            intermediate_count++;
        }

        x_start += x_inc;
        y_start += y_inc;

        vertexoffset_high += vertexoffset_increment;
        vertexoffset_low  += vertexoffset_increment;

        u2 += u_inc;
        v2 += v_inc;
    }

    if (indexcount >= 3)
    {
        IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_UV1 | VTX_UV2, (float*)Vertices,
            (int)vertexcount, Indices, indexcount);
    }
}
