// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "mtxlib.h"
#include "vector_math.h"
#include "get_distance.h"
#include "landscape_collider.h"
#include <iostream>

#define CONTACT(p,skip) ((dContactGeom*) (((char*)p) + (skip)))

int dLandscapeClass = -1;

// ============================================================================
// data for the landscape
// ============================================================================

struct dxLandscape
{
    float *landscape_heightfield;   // Pointer to the heightfield
    int landscape_width;            // width of landscape heightfield
    int landscape_height;           // height of landscape along z axis.
};

inline float getHeight(float *heightfield, int pitch, int i, int j)
{
    return *(heightfield + i + j * pitch);
}

void print_vector3(char *s, vector3 v)
{
    std::cout << s << " : " << " X: " << v.x << " Y: " << v.y << " Z: " << v.z << std::endl;
}

// Checks if a point (x, y, z) already exists in the contact array.
bool alreadyInContacts(dContactGeom* contact, int num_contacts, vector3 point, int skip)
{
    for (int i = 0; i < num_contacts; i++)
    {
        float x = CONTACT(contact, i * skip)->pos[0];
        float y = CONTACT(contact, i * skip)->pos[1];
        float z = CONTACT(contact, i * skip)->pos[2];

        float dist = get_distance(point, x, y, z);
        if (dist < 0.15) return true;
    }

    return false;
}

// Checks for collision between a Landscape and a Sphere
int dCollideLandscapeSphere (dGeomID o1, dGeomID o2, int flags, dContactGeom* contact, int skip)
{
    // Get the centre & radius of the sphere
    const dReal* centre = dGeomGetPosition(o2);
    dReal radius = dGeomSphereGetRadius(o2);

    vector3 centre_v(centre[0], centre[1], centre[2]);

    dReal aabb2[6];
    dGeomGetAABB (o2, aabb2);

    int left    = (int)floor(aabb2[0]);
    int right   = (int)ceil(aabb2[1]);

    int top     = (int)floor(aabb2[4]);
    int bottom  = (int)ceil(aabb2[5]);

    int num_contacts = 0;

    dxLandscape* l = (dxLandscape*) dGeomGetClassData(o1);
    float* heightfield = l->landscape_heightfield;
    int pitch = l->landscape_width;

    struct collision_info
    {
        //  v1--------v2
        //    |     /|
        //    |   /  |
        //    | /    |
        //  v3|-------v4

        // v1,v2,v3 is nw (north-west) face
        // v2,v4,v3 is se face

        bool hit_nw_face;   // used to indicate that nw face is in contact with the sphere.
        bool hit_se_face;   // used to indicate that se face is in contact with the sphere.
    };

    collision_info hits[128][128];  // Maximum sphere size is 128 metres in diameter
                                    // (or 128 units of whatever 1.0 units is in length).

    int j;

    for (j = top; j <= bottom; j++)
    {
        for (int i = left; i <= right; i++)
        {
            if ((i > 0) && (i < l->landscape_width - 1) && (j > 0) && (j < l->landscape_height - 1))
            {
                vector3 v1(i    , getHeight(heightfield, pitch, i    , j    ), j    );
                vector3 v2(i + 1, getHeight(heightfield, pitch, i + 1, j    ), j    );
                vector3 v3(i    , getHeight(heightfield, pitch, i    , j + 1), j + 1);
                vector3 v4(i + 1, getHeight(heightfield, pitch, i + 1, j + 1), j + 1);

                vector3 contact1;

                bool hit = closestPointOnTriangleSurface(v1, v2, v3, centre_v, contact1, radius);

                hits[i - left][j - top].hit_nw_face = hit;

                float distance = get_distance(contact1, centre[0], centre[1], centre[2]);

                if ((distance < radius) && hit && (num_contacts < 8) &&
                    !alreadyInContacts(contact, num_contacts, contact1, skip))
                {
                    float x = contact1.x;
                    float y = contact1.y;
                    float z = contact1.z;

                    vector3 normal(x - centre[0], y - centre[1], z - centre[2]);
                    normal.normalize();

                    CONTACT(contact, num_contacts * skip)->normal[0] = normal.x;
                    CONTACT(contact, num_contacts * skip)->normal[1] = normal.y;
                    CONTACT(contact, num_contacts * skip)->normal[2] = normal.z;

                    vector3 contact_point(x,y,z);

                    CONTACT(contact, num_contacts * skip)->pos[0] = contact_point.x;
                    CONTACT(contact, num_contacts * skip)->pos[1] = contact_point.y;
                    CONTACT(contact, num_contacts * skip)->pos[2] = contact_point.z;

                    CONTACT(contact, num_contacts * skip)->depth = radius - distance;

                    CONTACT(contact, num_contacts * skip)->g1 = o1;
                    CONTACT(contact, num_contacts * skip)->g2 = o2;

                    num_contacts ++;
                }

                hit = closestPointOnTriangleSurface(v2, v4, v3, centre_v, contact1, radius);

                hits[i - left][j - top].hit_se_face = hit;

                distance = get_distance(contact1, centre[0], centre[1], centre[2]);

                if ((distance < radius) && hit && (num_contacts < 8) &&
                    !alreadyInContacts(contact, num_contacts, contact1, skip))
                {
                    float x = contact1.x;
                    float y = contact1.y;
                    float z = contact1.z;

                    vector3 normal(x - centre[0], y - centre[1], z - centre[2]);
                    normal.normalize();

                    CONTACT(contact, num_contacts * skip)->normal[0] = normal.x;
                    CONTACT(contact, num_contacts * skip)->normal[1] = normal.y;
                    CONTACT(contact, num_contacts * skip)->normal[2] = normal.z;


                    vector3 contact_point(x,y,z);

                    CONTACT(contact, num_contacts * skip)->pos[0] = contact_point.x;
                    CONTACT(contact, num_contacts * skip)->pos[1] = contact_point.y;
                    CONTACT(contact, num_contacts * skip)->pos[2] = contact_point.z;

                    CONTACT(contact, num_contacts * skip)->depth = radius - distance;

                    CONTACT(contact, num_contacts * skip)->g1 = o1;
                    CONTACT(contact, num_contacts * skip)->g2 = o2;

                    num_contacts ++;
                }

            }
        }
    }

    for (j = top; j <= bottom; j++)
    {
        for (int i = left; i <= right; i++)
        {
            if ((i > 1) && (i < l->landscape_width - 1) && (j > 1) && (j < l->landscape_height - 1))
            {
                vector3 v1(i    , getHeight(heightfield, pitch, i    , j    ), j    );
                vector3 v2(i + 1, getHeight(heightfield, pitch, i + 1, j    ), j    );
                vector3 v3(i    , getHeight(heightfield, pitch, i    , j + 1), j + 1);
                vector3 v4(i + 1, getHeight(heightfield, pitch, i + 1, j + 1), j + 1);

                vector3 contact1;

                if (!hits[i - left      ][j - top       ].hit_nw_face &&
                    !hits[i - left      ][j - top       ].hit_se_face &&
                    ((i != left) && !hits[i - left - 1  ][j - top       ].hit_se_face) &&
                    ((j != top ) && !hits[i - left      ][j - top - 1   ].hit_se_face))
                {
                    contact1 = closestPointOnTriangle(v1, v2, v3, centre_v);

                    float distance = get_distance(contact1, centre[0], centre[1], centre[2]);

                    if ((distance < radius) && (num_contacts < 8) &&
                        !alreadyInContacts(contact, num_contacts, contact1, skip))
                    {
                        float x = contact1.x;
                        float y = contact1.y;
                        float z = contact1.z;

                        vector3 normal(x - centre[0], y - centre[1], z - centre[2]);
                        normal.normalize();

                        CONTACT(contact, num_contacts * skip)->normal[0] = normal.x;
                        CONTACT(contact, num_contacts * skip)->normal[1] = normal.y;
                        CONTACT(contact, num_contacts * skip)->normal[2] = normal.z;

                        vector3 contact_point(x,y,z);

                        CONTACT(contact, num_contacts * skip)->pos[0] = contact_point.x;
                        CONTACT(contact, num_contacts * skip)->pos[1] = contact_point.y;
                        CONTACT(contact, num_contacts * skip)->pos[2] = contact_point.z;

                        CONTACT(contact, num_contacts * skip)->depth = radius - distance;

                        CONTACT(contact, num_contacts * skip)->g1 = o1;
                        CONTACT(contact, num_contacts * skip)->g2 = o2;

                        num_contacts ++;
                    }

                    contact1 = closestPointOnTriangle(v2, v4, v3, centre_v);

                    distance = get_distance(contact1, centre[0], centre[1], centre[2]);

                    if ((distance < radius) && (num_contacts < 8) &&
                        !alreadyInContacts(contact, num_contacts, contact1, skip))
                    {

                        float x = contact1.x;
                        float y = contact1.y;
                        float z = contact1.z;

                        vector3 normal(x - centre[0], y - centre[1], z - centre[2]);
                        normal.normalize();

                        CONTACT(contact, num_contacts * skip)->normal[0] = normal.x;
                        CONTACT(contact, num_contacts * skip)->normal[1] = normal.y;
                        CONTACT(contact, num_contacts * skip)->normal[2] = normal.z;

                        vector3 contact_point(x,y,z);

                        CONTACT(contact, num_contacts * skip)->pos[0] = contact_point.x;
                        CONTACT(contact, num_contacts * skip)->pos[1] = contact_point.y;
                        CONTACT(contact, num_contacts * skip)->pos[2] = contact_point.z;

                        CONTACT(contact, num_contacts * skip)->depth = radius - distance;

                        CONTACT(contact, num_contacts * skip)->g1 = o1;
                        CONTACT(contact, num_contacts * skip)->g2 = o2;

                        num_contacts ++;
                    }
                }
            }
        }
    }

    return num_contacts;
}

// Checks for collision between a Landscape and a CCylinder
int dCollideLandscapeCCylinder(dGeomID o1, dGeomID o2, int flags, dContactGeom* contact, int skip)
{
    int num_contacts = 0;




    return num_contacts;
}

// Checks for collision between a Landscape and a Box
int dCollideLandscapeBox (dGeomID o1, dGeomID o2, int flags, dContactGeom* contact, int skip)
{
    int num_contacts = 0;

    // Get the side lengths of the box
    dVector3 box_sides;
    dGeomBoxGetLengths(o2, box_sides);

    // Get the rotation of the box
    const dReal *R = dGeomGetRotation(o2);

    // Get the position of the box
    const dReal *P = dGeomGetPosition(o2);

    // Get the 8 points of the box
    vector3 points[8];

    points[0].x = R[0] * -box_sides[0]/2.0 + R[1] * -box_sides[1]/2.0 + R[2] * -box_sides[2]/2.0;
    points[0].y = R[4] * -box_sides[0]/2.0 + R[5] * -box_sides[1]/2.0 + R[6] * -box_sides[2]/2.0;
    points[0].z = R[8] * -box_sides[0]/2.0 + R[9] * -box_sides[1]/2.0 + R[10]* -box_sides[2]/2.0;

    points[1].x = R[0] *  box_sides[0]/2.0 + R[1] * -box_sides[1]/2.0 + R[2]  * -box_sides[2]/2.0;
    points[1].y = R[4] *  box_sides[0]/2.0 + R[5] * -box_sides[1]/2.0 + R[6]  * -box_sides[2]/2.0;
    points[1].z = R[8] *  box_sides[0]/2.0 + R[9] * -box_sides[1]/2.0 + R[10] * -box_sides[2]/2.0;

    points[2].x = R[0] *  box_sides[0]/2.0 + R[1] * box_sides[1]/2.0 + R[2]  * -box_sides[2]/2.0;
    points[2].y = R[4] *  box_sides[0]/2.0 + R[5] * box_sides[1]/2.0 + R[6]  * -box_sides[2]/2.0;
    points[2].z = R[8] *  box_sides[0]/2.0 + R[9] * box_sides[1]/2.0 + R[10] * -box_sides[2]/2.0;

    points[3].x = R[0] *  -box_sides[0]/2.0 + R[1] * box_sides[1]/2.0 + R[2]  * -box_sides[2]/2.0;
    points[3].y = R[4] *  -box_sides[0]/2.0 + R[5] * box_sides[1]/2.0 + R[6]  * -box_sides[2]/2.0;
    points[3].z = R[8] *  -box_sides[0]/2.0 + R[9] * box_sides[1]/2.0 + R[10] * -box_sides[2]/2.0;

    points[4].x = R[0] *  -box_sides[0]/2.0 + R[1] * -box_sides[1]/2.0 + R[2]  * box_sides[2]/2.0;
    points[4].y = R[4] *  -box_sides[0]/2.0 + R[5] * -box_sides[1]/2.0 + R[6]  * box_sides[2]/2.0;
    points[4].z = R[8] *  -box_sides[0]/2.0 + R[9] * -box_sides[1]/2.0 + R[10] * box_sides[2]/2.0;

    points[5].x = R[0] *  box_sides[0]/2.0 + R[1] * -box_sides[1]/2.0 + R[2]  * box_sides[2]/2.0;
    points[5].y = R[4] *  box_sides[0]/2.0 + R[5] * -box_sides[1]/2.0 + R[6]  * box_sides[2]/2.0;
    points[5].z = R[8] *  box_sides[0]/2.0 + R[9] * -box_sides[1]/2.0 + R[10] * box_sides[2]/2.0;

    points[6].x = R[0] *  box_sides[0]/2.0 + R[1] * box_sides[1]/2.0 + R[2]  * box_sides[2]/2.0;
    points[6].y = R[4] *  box_sides[0]/2.0 + R[5] * box_sides[1]/2.0 + R[6]  * box_sides[2]/2.0;
    points[6].z = R[8] *  box_sides[0]/2.0 + R[9] * box_sides[1]/2.0 + R[10] * box_sides[2]/2.0;

    points[7].x = R[0] *  -box_sides[0]/2.0 + R[1] * box_sides[1]/2.0 + R[2]  * box_sides[2]/2.0;
    points[7].y = R[4] *  -box_sides[0]/2.0 + R[5] * box_sides[1]/2.0 + R[6]  * box_sides[2]/2.0;
    points[7].z = R[8] *  -box_sides[0]/2.0 + R[9] * box_sides[1]/2.0 + R[10] * box_sides[2]/2.0;

    dxLandscape* l = (dxLandscape*) dGeomGetClassData(o1);
    float* heightfield = l->landscape_heightfield;
    int pitch = l->landscape_width;

    // Check all the points against the terrain
    for (int p = 0; p < 8; p++)
    {
        // Add the position of the box to each point
        points[p].x += P[0];
        points[p].y += P[1];
        points[p].z += P[2];

        int i = (int)points[p].x;
        int j = (int)points[p].z;

        float land_height = getHeight(heightfield, pitch, i, j);

        vector3 v1(i    , getHeight(heightfield, pitch, i    , j    ), j    );
        vector3 v2(i + 1, getHeight(heightfield, pitch, i + 1, j    ), j    );
        vector3 v3(i    , getHeight(heightfield, pitch, i    , j + 1), j + 1);

        vector3 normal = CrossProduct(v2 - v1, v3 - v1).normalize();

        float depth = land_height - points[p].y;

        if (depth >= 0.0)
        {
            CONTACT(contact, num_contacts * skip)->normal[0] = normal.x;
            CONTACT(contact, num_contacts * skip)->normal[1] = normal.y;
            CONTACT(contact, num_contacts * skip)->normal[2] = normal.z;

            CONTACT(contact, num_contacts * skip)->pos[0] = points[p].x;
            CONTACT(contact, num_contacts * skip)->pos[1] = points[p].y;
            CONTACT(contact, num_contacts * skip)->pos[2] = points[p].z;

            CONTACT(contact, num_contacts * skip)->depth = depth;

            CONTACT(contact, num_contacts * skip)->g1 = o1;
            CONTACT(contact, num_contacts * skip)->g2 = o2;

            num_contacts++;

            if (num_contacts > 3) return num_contacts;
        }
    }

    return num_contacts;
}


// Checks for collision between a Landscape and an AABB.
int dCollideLandscapeAABB (dGeomID o1, dGeomID o2, dReal aabb2[6])
{
    // This function checks for contact between the Landscape and an AABB.
    // It checks a larger region of terrain than strictly necessary and
    // will sometimes return true even if there is no collision, but will never
    // return false if there is a collision.
    // If the AABB is below the terrain it will also return true.

    int left    = (int)floor(aabb2[0]);
    int right   = (int)ceil(aabb2[1]);

    int top     = (int)floor(aabb2[4]);
    int bottom  = (int)ceil(aabb2[5]);

    int collide = 0;

    dxLandscape *l = (dxLandscape*) dGeomGetClassData(o1);
    float *heightfield = l->landscape_heightfield;
    int pitch = l->landscape_width;

    for (int j = top; j <= bottom; j++)
    {
        for (int i = left; i <= right; i++)
        {
            if ((i > 0) && (i < l->landscape_width) && (j > 0) && (j < l->landscape_height))
            {
                // If the height of this point in the heightfield is greater than the
                // lower edge of the aabb ...
                if (getHeight(heightfield, pitch, i, j) > aabb2[2]) collide = 1;
            }
        }
    }
    return collide;
}

static dColliderFn* dLandscapeColliderFn (int num)
{
    if (num == dSphereClass)    return (dColliderFn *) &dCollideLandscapeSphere;
    if (num == dCCylinderClass) return (dColliderFn *) &dCollideLandscapeCCylinder;
    if (num == dBoxClass)       return (dColliderFn *) &dCollideLandscapeBox;
    return 0;
}

static void dLandscapeAABB (dxGeom *geom, dReal aabb[6])
{
    // If you have mostly airborne objects or break your terrain up
    // into patches of individual terrain collision geometries you
    // may wish to compute a tighter bounding box.
    aabb[0] = 0.0;
    aabb[1] = 513.0;
    aabb[2] = 0.0;
    aabb[3] = 64.0;
    aabb[4] = 0.0;
    aabb[5] = 513.0;
}

// Heightfield is an already allocated array width * height * sizeof(float) bytes in size.
dxGeom *dCreateLandscape (dSpaceID space, float *heightfield, int width, int height)
{
    dAASSERT (width  > 0);
    dAASSERT (height > 0);

    if (dLandscapeClass == -1)
    {
        dGeomClass c;
        c.bytes = sizeof (dxLandscape);
        c.collider = &dLandscapeColliderFn;
        c.aabb = &dLandscapeAABB;
        c.aabb_test = &dCollideLandscapeAABB;
        c.dtor = 0;
        dLandscapeClass = dCreateGeomClass (&c);
    }

    dxGeom *g = dCreateGeom (dLandscapeClass);
    if (space) dSpaceAdd (space,g);
    dxLandscape *l = (dxLandscape*) dGeomGetClassData(g);
    l->landscape_heightfield   = heightfield;
    l->landscape_width         = width;
    l->landscape_height        = height;
    return g;
}
