// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "object_renderer.h"
#include "get_distance.h"
#include <iostream>
#include <algorithm>

ObjectRenderer::ObjectRenderer()
{
    this->IGS = NULL;
}

ObjectRenderer::~ObjectRenderer()
{

}

void ObjectRenderer::Begin(IGraphicsSystem *IGS, vector3& camera, vector3& light_vector, matrix44 view_matrix)
{
    this->IGS = IGS;
    this->camera = camera;
    this->light_vector = light_vector;
    this->view_matrix = view_matrix;

    transparent_group.clear();
    model_group.clear();
}

// Adds a model to the scene
void ObjectRenderer::AddModel(IModel* model, matrix44 transformation_model, float time, ANIMATION_INSTANCE_HANDLE instance)
{
    // Determine the distance of the model from the camera
    vector3 object_centre = transform(vector3(0.0f, 0.0f, 0.0f), transformation_model);
    float distance = get_distance(camera, object_centre);

    // Calculate the bounding sphere size on screen.
    float radius = model->GetRadiusOfBoundingSphere();
    float projected_sphere_size = (radius * radius) / distance;

    Model m;
    m.model             = model;
    m.model_index       = model_group.size();
    m.parent_index      = -1;
    m.parent            = false;
    m.anim_time         = time;
    m.anim_updated      = false;
    m.anim_instance     = instance;
    m.sphere_size       = projected_sphere_size;
    m.distance          = distance;
    m.transformation    = transformation_model;

    model_group.push_back(m);
}

// Adds a group of models to the scene. This way their shadows fade together.
void ObjectRenderer::AddModelGroup(std::vector<IModel*> models, std::vector<matrix44> transformations, int parent_index)
{
    int new_parent_index = model_group.size() + parent_index;

    // Loop through all the objects
    for (unsigned int m = 0; m < models.size(); m++)
    {
        AddModel(models[m], transformations[m]);

        model_group[model_group.size() - 1].parent_index = new_parent_index;

        if (parent_index == (int)m)
        {
            model_group[model_group.size() - 1].parent = true;
            model_group[model_group.size() - 1].parent_index = -1;
        }
    }
}

// Adds a particle to the group of particles to be sorted and rendered.
void ObjectRenderer::AddParticle(float size, float opacity, blend_mode mode, TEXTURE_HANDLE texture, vector3 position)
{
    // from (http://nate.scuzzy.net/docs/billboard/)
    // Get the transpose of the view matrix
    matrix44 transposed = TransposeMatrix44(view_matrix);
    vector3 x = vector3(transposed.col[0].x, transposed.col[0].y, transposed.col[0].z);
    vector3 y = vector3(transposed.col[1].x, transposed.col[1].y, transposed.col[1].z);

    VERTEX_XYZ_COLOUR_UV1 vertices[4];
    vertices[0].u = 0.0f;
    vertices[0].v = 1.0f;
    vertices[0].colour.r = 1.0f;
    vertices[0].colour.g = 1.0f;
    vertices[0].colour.b = 1.0f;

    vertices[1].u = 0.0f;
    vertices[1].v = 0.0f;
    vertices[1].colour.r = 1.0f;
    vertices[1].colour.g = 1.0f;
    vertices[1].colour.b = 1.0f;

    vertices[2].u = 1.0f;
    vertices[2].v = 0.0f;

    vertices[2].colour.r = 1.0f;
    vertices[2].colour.g = 1.0f;
    vertices[2].colour.b = 1.0f;

    vertices[3].u = 1.0f;
    vertices[3].v = 1.0f;
    vertices[3].colour.r = 1.0f;
    vertices[3].colour.g = 1.0f;
    vertices[3].colour.b = 1.0f;

    vector3 a = position + ((-x - y) * size);
    vector3 b = position + ((-x + y) * size);
    vector3 c = position + (( x + y) * size);
    vector3 d = position + (( x - y) * size);

    for (int i = 0; i < 4; i++)
    {
        vertices[i].colour.a = opacity;
    }

    vertices[0].point = a;
    vertices[1].point = b;
    vertices[2].point = c;
    vertices[3].point = d;

    TransparentPolygon p1, p2;

    float distance = get_distance(camera, position);

    memcpy(p1.vertices, vertices, 3 * sizeof(VERTEX_XYZ_COLOUR_UV1));

    p1.distance = distance;
    p1.texture = texture;
    p1.mode = mode;

    p2.vertices[0] = vertices[2];
    p2.vertices[1] = vertices[3];
    p2.vertices[2] = vertices[0];
    p2.distance = distance;
    p2.texture = texture;
    p2.mode = mode;

    transparent_group.push_back(p1);
    transparent_group.push_back(p2);
}

// Renders all the objects in the scene
void ObjectRenderer::RenderObjects()
{
    // Loop through all the objects
    std::vector<Model>::iterator itor;
    for (itor = model_group.begin(); itor != model_group.end(); itor++)
    {
        IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, itor->transformation);

        // Determine if the object is within the frustum
        if (itor->model->IsWithinFrustum(itor->transformation))
        {
            if (itor->anim_instance != NULL)
            {
                itor->model->Update(itor->anim_time, itor->anim_instance);
                itor->anim_updated = true;
            }

            itor->model->Render(itor->anim_instance);
        }
    }
}

// Render all the shadows of the objects in the scene
void ObjectRenderer::RenderShadows()
{
    float clip_y = -0.1f;

    IGS->PrepareStencil();

    // Loop through all the objects
    std::vector<Model>::iterator itor;
    for (itor = model_group.begin(); itor != model_group.end(); itor++)
    {
        //float fade_range_near = 500.0f * itor->sphere_size;

        //if (itor->distance <= fade_range_near)
        {
            if (itor->parent_index != -1)
            {


            }
            // Determine whether or not the shadow is visible
            else if (itor->model->IsShadowWithinFrustum(itor->transformation, light_vector))
            {
                if (itor->anim_instance != NULL && itor->anim_updated == false)
                {
                    itor->model->Update(itor->anim_time, itor->anim_instance);
                    itor->anim_updated = true;
                }

                // Determine whether or not the camera is in the shadow volume
                shadow_mode mode = Z_PASS_SHADOWS;

                if (itor->model->IsCameraWithinShadow(itor->transformation, camera, light_vector))
                {
                    mode = Z_FAIL_SHADOWS;
                }

                itor->model->cast_shadow(itor->transformation, light_vector, itor->anim_instance, mode, clip_y);

                if (itor->parent == true)
                {
                    std::vector<Model>::iterator itor2;
                    for (itor2 = model_group.begin(); itor2 != model_group.end(); itor2++)
                    {
                        if (itor2->parent_index == itor->model_index)
                        {
                            itor2->model->cast_shadow(itor2->transformation, light_vector, itor2->anim_instance, mode, clip_y);
                        }
                    }
                }
            }
        }
    }
    /*
    IGS->StencilShadowRectangle(COLOUR(0.0f, 0.0f, 0.0f, 0.5f));

    // Loop through all the objects again
    for (itor = model_group.begin(); itor != model_group.end(); itor++)
    {
        float fade_range_near = 500.0f * itor->sphere_size;
        float fade_range_far  = 500.0f * itor->sphere_size + 20.0f;
        float fade_difference = fade_range_far - fade_range_near;

        if (itor->distance > fade_range_near && itor->distance <= fade_range_far)
        {
            if (itor->parent_index != -1)
            {


            }
            // Determine whether or not the shadow is visible
            else if (itor->model->IsShadowWithinFrustum(itor->transformation, light_vector))
            {
                if (itor->anim_instance != NULL && itor->anim_updated == false)
                {
                    itor->model->Update(itor->anim_time, itor->anim_instance);
                    itor->anim_updated = true;
                }

                // Determine whether or not the camera is in the shadow volume
                shadow_mode mode = Z_PASS_SHADOWS;

                if (itor->model->IsCameraWithinShadow(itor->transformation, camera, light_vector))
                {
                    mode = Z_FAIL_SHADOWS;
                }

                IGS->ClearStencil();
                IGS->PrepareStencil();

                itor->model->cast_shadow(itor->transformation, light_vector, itor->anim_instance, mode, clip_y);

                if (itor->parent == true)
                {
                    std::vector<Model>::iterator itor2;
                    for (itor2 = model_group.begin(); itor2 != model_group.end(); itor2++)
                    {
                        if (itor2->parent_index == itor->model_index)
                        {
                            itor2->model->cast_shadow(itor2->transformation, light_vector, itor2->anim_instance, mode, clip_y);
                        }
                    }
                }

                float alpha = ((fade_difference - (itor->distance - fade_range_near)) / fade_difference) / 2.0f;
                IGS->StencilShadowRectangle(COLOUR(0.0f, 0.0f, 0.0f, alpha));
            }
        }
    }
    */
}

// Renders all the particles
void ObjectRenderer::RenderTransparentObjects()
{
    // Check if the number of particles is 0 and return if so.
    if (transparent_group.size() == 0) return;

    // Sort the transparent polygons by distance
    std::sort(transparent_group.begin(), transparent_group.end());

    IGS->EnableLighting(false);
    IGS->SetTextureWrappingMode(0, CLAMP_TEXTURE);
    IGS->SetCullMode(CULL_NONE);
    IGS->SetDepthBufferMode(READ_ONLY);

    // Set the object to world transform to the identity
    matrix44 identity = IdentityMatrix44();
    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, identity);

    // Loop through all the transparent polygons and render them
    std::vector<TransparentPolygon>::iterator itor;
    for (itor = transparent_group.begin(); itor != transparent_group.end(); itor++)
    {
        // Make sure the correct texture is Set.
        IGS->SetTexture(0, itor->texture);
        // Make sure the correct blend mode is set.
        IGS->SetBlendMode(itor->mode);

        IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_COLOUR | VTX_UV1, (float*)itor->vertices, 3);
    }

    IGS->SetBlendMode(BLEND_DISABLE);
    IGS->SetDepthBufferMode(READ_AND_WRITE);
}

