// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef RENDER_OBJECTS_H
#define RENDER_OBJECTS_H

#include "transform.h"
#include "graphicssystem.h"
#include "particle_system.h"
#include "model.h"

class ObjectRenderer
{
public:
    //! Returns a reference to the one and only instance of this class.
    static ObjectRenderer& Instance()
    {
        static ObjectRenderer Obj;
        return Obj;
    }

    ~ObjectRenderer();

    // Call this once per frame before calling any of the other functions.
    void Begin(IGraphicsSystem *IGS, vector3& camera, vector3& light_vector, matrix44 view_matrix);

    // Adds a model to the scene
    void AddModel(IModel* model, matrix44 transformation_model, float time = 0.0f, ANIMATION_INSTANCE_HANDLE instance = NULL);

    // Adds a group of models to the scene. This way their shadows fade together.
    void AddModelGroup(std::vector<IModel*> models, std::vector<matrix44> transformations, int parent_index);

    // Adds a particle to the group of particles to be sorted and rendered.
    void AddParticle(float size, float opacity, blend_mode mode, TEXTURE_HANDLE texture, vector3 position);

    // Renders all the objects in the scene
    void RenderObjects();

    // Render all the shadows of the objects in the scene
    void RenderShadows();

    // Renders all the particles
    void RenderTransparentObjects();

private:

    // Constructor is private in this singleton
    ObjectRenderer();

    // A pointer to the rendering interface
    IGraphicsSystem *IGS;

    // The camera position
    vector3 camera;

    // The light vector
    vector3 light_vector;

    //  The view matrix
    matrix44 view_matrix;

    // A structure holding all the info needed to render a transparent triangle polygon
    class TransparentPolygon
    {
    public:
        VERTEX_XYZ_COLOUR_UV1 vertices[3];  // The 3 vertices of the trinagle.
        float distance;                     // The distance of the triangle from the camera
        TEXTURE_HANDLE texture;             // Texture used to render the particle.
        blend_mode mode;                    // Determines the blend mode

        // comparison operator. this will be called by std::sort and will sort them by distance
        bool operator<(const TransparentPolygon& other) const
        {
            return (this->distance > other.distance);
        }
    };

    // Our group of particles to be sorted.
    std::vector<TransparentPolygon> transparent_group;

    class Model
    {
    public:
        IModel* model;
        int model_index;
        int parent_index;
        bool parent;
        float anim_time;
        bool anim_updated;
        ANIMATION_INSTANCE_HANDLE anim_instance;
        float sphere_size;
        float distance;
        matrix44 transformation;
    };

    // The group of models
    std::vector<Model> model_group;
};

#endif // RENDER_OBJECTS_H

