// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include "SDL_thread.h"
#include <iostream>
#include "ode_thread.h"
#include "body_user_data_ids.h"
#include "gettime.h"

SDL_mutex* ode_mutex;
SDL_Thread* ode_thread;

OdeThread::OdeThread()
{
    // Initialise ODE
    world = dWorldCreate();
    space = dHashSpaceCreate(0);
    contactgroup = dJointGroupCreate(0);
    dWorldSetGravity (world, 0, -0.8, 0);
    dWorldSetCFM (world, 1e-1);

    thread_exit = false;

    ode_mutex = SDL_CreateMutex();
}

OdeThread::~OdeThread()
{
    dJointGroupDestroy (contactgroup);
    dSpaceDestroy (space);
    dWorldDestroy (world);
}

// Starts the thread
void OdeThread::start_thread()
{
    ode_thread = SDL_CreateThread(ode_thread_function, this);
    if (ode_thread == NULL)
    {
        std::cout << "error creating thread." << std::endl;
    }
}

// Stops the thread
void OdeThread::stop_thread()
{
    ode_mutex_lock();

    thread_exit = true;

    ode_mutex_unlock();

    SDL_WaitThread(ode_thread, NULL);
}

// Returns the ODE world object
dWorldID OdeThread::getWorld()
{
    ode_mutex_lock();

    dWorldID temp = world;

    ode_mutex_unlock();

    return temp;
}

// Returns the ODE space object
dSpaceID OdeThread::getSpace()
{
    ode_mutex_lock();

    dSpaceID temp = space;

    ode_mutex_unlock();

    return temp;
}

void OdeThread::ode_mutex_lock()
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    if (SDL_mutexP(ode_mutex) == -1)
    {
        std::cout << "Couldn't lock mutex." << std::endl;
    }
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
}

void OdeThread::ode_mutex_unlock()
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    if (SDL_mutexV(ode_mutex) == -1)
    {
        std::cout << "Couldn't lock mutex." << std::endl;
    }
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
}

// this is called by dSpaceCollide when two objects in space are potentially colliding.
void nearCallback(OdeThread* ot, dGeomID o1, dGeomID o2)
{
    if (dGeomIsSpace(o1) || dGeomIsSpace(o2))
    {
        // colliding a space with something
        dSpaceCollide2(o1, o2, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);

        // collide all geoms internal to the space(s)
        if (dGeomIsSpace (o1)) dSpaceCollide((dSpaceID)o1, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);
        if (dGeomIsSpace (o2)) dSpaceCollide((dSpaceID)o2, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);
    }
    else
    {
        // exit without doing anything if the two bodies are connected by a joint
        dBodyID b1 = dGeomGetBody(o1);
        dBodyID b2 = dGeomGetBody(o2);
        if (b1 && b2 && dAreConnected(b1,b2)) return;

        dContact contact[8];   // up to 5 contacts
        if (int numc = dCollide (o1, o2, 8, &contact[0].geom, sizeof(dContact)))
        {
            for (int i=0; i < numc; i++)
            {
                if ((b1 != 0 && dBodyGetData(b1) == (void *)BODY_BUGGY) ||
                    (b2 != 0 && dBodyGetData(b2) == (void *)BODY_BUGGY))
                {
                    contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM;
                    contact[i].surface.mu = dInfinity;
                    contact[i].surface.slip1 = 1.0;
                    contact[i].surface.slip2 = 0.5;
                    contact[i].surface.soft_erp = 0.5;
                    contact[i].surface.soft_cfm = 0.3;
                }
                if ((b1 != 0 && dBodyGetData(b1) == (void *)BODY_PLAYER) ||
                    (b2 != 0 && dBodyGetData(b2) == (void *)BODY_PLAYER))
                {
                    contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM;
                    contact[i].surface.mu = dInfinity;
                    contact[i].surface.slip1 = 1.0;
                    contact[i].surface.slip2 = 0.5;
                    contact[i].surface.soft_erp = 0.2;
                    contact[i].surface.soft_cfm = 1.0;
                }
                else // default
                {
                    contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM;
                    contact[i].surface.mu = dInfinity;
                    contact[i].surface.slip1 = 1.0;
                    contact[i].surface.slip2 = 0.5;
                    contact[i].surface.soft_erp = 0.5;
                    contact[i].surface.soft_cfm = 0.3;
                }

                dJointID c = dJointCreateContact (ot->world, ot->contactgroup, contact+i);

                dJointAttach (c,b1,b2);
            }
        }
    }
}



void OdeThread::Update()
{
    const int updates_per_second = 80;

    int update_time = 1000 / updates_per_second;
    if (update_time == 0) update_time = 1;

    static long start = GetTime();

    static long remaining_time = 0;

    remaining_time += GetTime() - start;
    start = GetTime();

    if (remaining_time > 250)
    {
        // More than a 250ms have elapsed.
        // This could be caused by the Operating System chugging away at some task or by the
        // user ALT+TABBING away, or by a v. slow framerate ( < 4 fps ). In any case we don't
        // update the physics engine.
        remaining_time = 0;
    }

    while (remaining_time > update_time)
    {
        dSpaceCollide(space, this, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);

        float step_size = 6.0f / float(updates_per_second);

        dWorldStep(world, step_size);

        // Remove all contact joints
        dJointGroupEmpty (contactgroup);

        remaining_time -= update_time;
    }
}

int ode_thread_function(void* void_ptr)
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    OdeThread* ot = (OdeThread*)void_ptr;

    ot->ode_mutex_lock();

    while (ot->thread_exit == false)
    {
        ot->Update();

        ot->ode_mutex_unlock();

        // Sleep for 5 milliseconds.
        SleepMS(5);

        ot->ode_mutex_lock();
    }

    ot->ode_mutex_unlock();
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
    return 0;
}

