// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "particle_system.h"
#include "object_renderer.h"
#include "gettime.h"
#include <iostream>

void ParticleSystem::CreateParticle(Particle& new_particle, long start_time)
{
    InternalParticle ip;

    ip.p = new_particle;
    ip.start_time = start_time;

    particles.push_back(ip);
}

void ParticleSystem::UpdateParticles()
{
    if (particles.size() == 0) return;

    long current_time = GetTime();

    // Loop through all the particles updating their position
    std::vector<InternalParticle>::iterator itor;
    for (itor = particles.begin(); itor != particles.end(); itor++)
    {
        long age = current_time - itor->start_time;

        itor->position = itor->p.initial_position + itor->p.velocity * (float)age / 1000.0f;
    }

    // Loop through all the particles updating their current appearance
    for (itor = particles.begin(); itor != particles.end(); itor++)
    {
        long age = current_time - itor->start_time;

        // dead ranges from 0.0 (newborn) to 1.0 (dead)
        float dead = (float)age / (float)itor->p.maximum_age;

        // calculate the current appearence by interpolating the initial and final appearances.
        float size_delta = (float)itor->p.final.size - (float)itor->p.initial.size;
        float opacity_delta = (float)itor->p.final.opacity - (float)itor->p.initial.opacity;

        itor->current_appearance.size = (float)itor->p.initial.size + (size_delta * dead);
        itor->current_appearance.opacity = (float)itor->p.initial.opacity + (opacity_delta * dead);
    }
}

void ParticleSystem::SubmitParticlesToParticleRenderer()
{
    if (particles.size() == 0) return;
    
    long current_time = GetTime();

    // Loop through all the particles and submit them to be rendered
    std::vector<InternalParticle>::iterator itor;
    for (itor = particles.begin(); itor != particles.end(); itor++)
    {
        long age = current_time - itor->start_time;

        if (age > itor->p.maximum_age)
        {
            // Don't render this particle
        }
        else
        {
            ObjectRenderer::Instance().AddParticle(itor->current_appearance.size,
                itor->current_appearance.opacity, itor->p.mode, itor->p.texture, itor->position);
        }
    }
}
